@@ -15,8 +15,11 @@ from typing import Set
 from typing import Tuple
 from typing import Union
 
+import pystac
 import yaml
+from deepdiff import DeepDiff
 from more_itertools import take
+from pystac import Catalog
 from pystac import Collection
 from pystac import Item
 from pystac import ItemCollection
@@ -94,6 +97,8 @@ class Context(Enum):
     COLLECTIONS = "Collections"
     ITEM_SEARCH_FILTER = "Item Search - Filter Ext"
     FEATURES_FILTER = "Features - Filter Ext"
+    CHILDREN = "Children Ext"
+    BROWSEABLE = "Browseable Ext"
 
     def __str__(self) -> str:
         return self.value
@@ -258,6 +263,35 @@ def supports(conforms_to: List[str], pattern: Pattern[str]) -> bool:
     return any(pattern.fullmatch(x) for x in conforms_to)
 
 
+def is_json_type(maybe_type: Optional[str]) -> bool:
+    return maybe_type is not None and (
+        maybe_type == "application/json" or maybe_type.startswith("application/json;")
+    )
+
+
+def is_geojson_type(maybe_type: Optional[str]) -> bool:
+    return maybe_type is not None and (
+        maybe_type == "application/geo+json"
+        or maybe_type.startswith("application/geo+json;")
+    )
+
+
+# def is_json_or_geojson_type(maybe_type: Optional[str]) -> bool:
+#     return maybe_type and (is_json_type(maybe_type) or is_geojson_type(maybe_type))
+
+
+def has_content_type(headers: Mapping[str, str], content_type: str) -> bool:
+    return headers.get("content-type", "").split(";")[0] == content_type
+
+
+def has_json_content_type(headers: Mapping[str, str]) -> bool:
+    return is_json_type(headers.get("content-type"))
+
+
+def has_geojson_content_type(headers: Mapping[str, str]) -> bool:
+    return is_geojson_type(headers.get("content-type"))
+
+
 def stac_validate(
     url: str,
     body: Optional[Dict[str, Any]],
@@ -304,8 +338,8 @@ def stac_check(
         errors += (
             f"[{context}] {method} {url} is not a valid STAC object: {linter.error_msg}"
         )
-    if linter.best_practices_msg:
-        warnings += f"[{context}] {method} {url} has these stac-check recommendations: {linter.best_practices_msg}"
+    if msgs := linter.best_practices_msg[1:]:  # first msg is a header
+        warnings += f"[{context}] {method} {url} has these stac-check recommendations: {''.join(msgs)}"
 
 
 def retrieve(
@@ -330,7 +364,7 @@ def retrieve(
 
     if resp.status_code != status_code:
         errors += (
-            f"[{context}] {method} {url} params={params} body={body}"
+            f"[{context}] {method} {url} params={params} body={json.dumps(body) if body else ''}"
             f" had unexpected status code {resp.status_code} instead of {status_code}: {additional}"
         )
 
@@ -344,7 +378,9 @@ def retrieve(
         elif not has_content_type(resp.headers, content_type):
             errors += f"[{context}] {method} {url} params={params} body={body} content-type header is not '{content_type}'"
 
-        if resp.headers.get("content-type", "").split(";")[0].endswith("json"):
+        if has_json_content_type(resp.headers) or has_geojson_content_type(
+            resp.headers
+        ):
             try:
                 return resp.status_code, resp.json(), resp.headers
             except json.decoder.JSONDecodeError:
@@ -362,7 +398,7 @@ def validate_core_landing_page_body(
     collection: Optional[str],
     geometry: Optional[str],
 ) -> bool:
-    if not has_content_type(headers, "application/json"):
+    if not has_json_content_type(headers):
         errors += (
             "CORE-1",
             "[Core] : Landing Page (/) response Content-Type header is not application/json",
@@ -450,10 +486,6 @@ def validate_core_landing_page_body(
     return True
 
 
-def has_content_type(headers: Mapping[str, str], content_type: str) -> bool:
-    return headers.get("content-type", "").split(";")[0] == content_type
-
-
 def validate_api(
     root_url: str,
     conformance_classes: List[str],
@@ -499,21 +531,15 @@ def validate_api(
 
     if "browseable" in conformance_classes:
         logger.info("Validating STAC API - Browseable conformance class.")
-        validate_browseable(landing_page_body, errors, warnings)
-    else:
-        logger.info("Skipping STAC API - Browseable conformance class.")
+        validate_browseable(landing_page_body, errors, warnings, r_session)
 
     if "children" in conformance_classes:
         logger.info("Validating STAC API - Children conformance class.")
-        validate_children(landing_page_body, errors, warnings)
-    else:
-        logger.info("Skipping STAC API - Children conformance class.")
+        validate_children(landing_page_body, errors, warnings, r_session)
 
     if "collections" in conformance_classes:
         logger.info("Validating STAC API - Collections conformance class.")
         validate_collections(landing_page_body, collection, errors, warnings, r_session)
-    else:
-        logger.info("Skipping STAC API - Collections conformance class.")
 
     conforms_to = landing_page_body.get("conformsTo", [])
 
@@ -529,8 +555,6 @@ def validate_api(
             errors,
             r_session,
         )
-    else:
-        logger.info("Skipping STAC API - Features conformance class.")
 
     if "item-search" in conformance_classes:
         logger.info("Validating STAC API - Item Search conformance class.")
@@ -545,8 +569,6 @@ def validate_api(
             conformance_classes=conformance_classes,
             r_session=r_session,
         )
-    else:
-        logger.info("Skipping STAC API - Item Search conformance class.")
 
     if not errors:
         try:
@@ -591,14 +613,14 @@ def validate_core(
     if not (root := link_by_rel(links, "root")):
         errors += "/ : Link[rel=root] must exist"
     else:
-        if root.get("type") != "application/json":
-            errors += "/ : Link[rel=root] type is not application/json"
+        if not is_json_type(root.get("type")):
+            errors += f"/ : Link[rel=root] type is not application/geo+json, instead {root.get('type')}"
 
     if not (_self := link_by_rel(links, "self")):
         warnings += "/ : Link[rel=self] must exist"
     else:
-        if _self.get("type") != "application/json":
-            errors += "/ : Link[rel=self] type is not application/json"
+        if not is_json_type(_self.get("type")):
+            errors += f"/ : Link[rel=self] type is not application/geo+json, instead {_self.get('type')}"
 
     if not (service_desc := link_by_rel(links, "service-desc")):
         errors += "/ : Link[rel=service-desc] must exist"
@@ -654,21 +676,132 @@ def validate_core(
             r_session=r_session,
         )
 
+    # this validates, among other things, that the child and item link relations reference
+    # valid STAC Catalogs, Collections, and/or Items
+    try:
+        list(take(1000, Catalog.from_dict(root_body).get_all_items()))
+    except pystac.errors.STACTypeError as e:
+        errors += (
+            f"[{Context.CORE}] Error while traversing Catalog child/item links to find Items: {e} "
+            "This can be reproduced with 'list(pystac.Catalog.from_file(root_url).get_all_items())'"
+        )
+
 
 def validate_browseable(
     root_body: Dict[str, Any],
     errors: Errors,
     warnings: Warnings,
+    r_session: Session,
 ) -> None:
-    logger.info("Browseable validation is not yet implemented.")
+    # child or item links exist in the root
+    child_links = links_by_rel(root_body.get("links"), "child")
+    item_links = links_by_rel(root_body.get("links"), "item")
+    if not (child_links or item_links):
+        errors += f"[{Context.BROWSEABLE}] /: Root catalog does not contain any child or item link relations"
+
+    # check that at least a few of the items that can be reached from child/item link relations
+    # can be found through search
+    try:
+        for item in take(10, Catalog.from_dict(root_body).get_all_items()):
+            if link := link_by_rel(root_body.get("links"), "search"):
+                _, body, _ = retrieve(
+                    Method.GET,
+                    link["href"],
+                    errors,
+                    Context.BROWSEABLE,
+                    params={"ids": item.id, "collections": item.collection},
+                    r_session=r_session,
+                )
+                if body and len(body.get("features", [])) != 1:
+                    errors += f"[{Context.BROWSEABLE}] /: Link[rel=children] must href /children"
+            else:
+                errors += (
+                    f"[{Context.BROWSEABLE}] /: Link[rel=search] could not be found"
+                )
+    except pystac.errors.STACTypeError as e:
+        errors += (
+            f"[{Context.BROWSEABLE}] Error while traversing Catalog child/item links to find Items: {e}. "
+            "This can be reproduced with 'pystac.Catalog.from_file(root_url).get_all_items()'"
+        )
 
 
 def validate_children(
     root_body: Dict[str, Any],
     errors: Errors,
     warnings: Warnings,
+    r_session: Session,
 ) -> None:
-    logger.info("Children validation is not yet implemented.")
+    children_link = link_by_rel(root_body.get("links"), "children")
+    if (
+        not children_link
+        or not children_link.get("href", "").endswith("/children")
+        or not is_json_type(children_link.get("type"))
+    ):
+        errors += f"[{Context.CHILDREN}] /: Link[rel=children] must href /children"
+        return
+
+    if not (children_href := children_link.get("href")):
+        errors += f"[{Context.CHILDREN}] /: Link[rel=children] missing href"
+    else:
+        _, children_body, resp_headers = retrieve(
+            Method.GET,
+            children_href,
+            errors,
+            Context.CHILDREN,
+            r_session=r_session,
+        )
+        if not children_body:
+            errors += f"[{Context.CHILDREN}] /children body was empty"
+            return
+
+        if not resp_headers or not has_json_content_type(resp_headers):
+            errors += f"[{Context.CHILDREN}] /children content-type header was not application/json"
+
+        if not (self_link := link_by_rel(children_body.get("links", []), "self")):
+            errors += f"[{Context.CHILDREN}] /children does not have self link"
+        elif children_link.get("href") != self_link.get("href"):
+            errors += (
+                f"[{Context.CHILDREN}] /children self link does not match requested url"
+            )
+
+        if not link_by_rel(children_body.get("links", []), "root"):
+            errors += f"[{Context.CHILDREN}] /children does not have root link"
+
+        # each child link in Landing Page must have an entry in children
+        child_links = links_by_rel(root_body.get("links"), "child")
+
+        child_link_bodies = []
+        for child_link in child_links:
+            if child_href := child_link.get("href"):
+                _, child_body, child_resp_headers = retrieve(
+                    Method.GET,
+                    child_href,
+                    errors,
+                    Context.CHILDREN,
+                    r_session=r_session,
+                )
+                child_link_bodies.append(child_body)
+            else:
+                errors += f"[{Context.CHILDREN}] child link {json.dumps(child_link)} missing href field"
+
+        child_links_vs_children_diff = DeepDiff(
+            child_link_bodies, children_body.get("children"), ignore_order=True
+        )
+        if iterable_item_removed := child_links_vs_children_diff.get(
+            "iterable_item_removed"
+        ):
+            errors += (
+                f"[{Context.CHILDREN}] /: child links contained these objects that /children does not: "
+                f"{json.dumps(iterable_item_removed)}"
+            )
+
+        if iterable_item_added := child_links_vs_children_diff.get(
+            "iterable_item_added"
+        ):
+            errors += (
+                f"[{Context.CHILDREN}] /: child links missing these objects that /children contains: "
+                f"{json.dumps(iterable_item_added)}"
+            )
 
 
 def validate_collections(
@@ -703,11 +836,7 @@ def validate_collections(
         if not body:
             errors += f"[{Context.COLLECTIONS}] /collections body was empty"
         else:
-            if (
-                not resp_headers
-                or resp_headers.get("content-type", "").split(";")[0]
-                != "application/json"
-            ):
+            if not resp_headers or not has_json_content_type(resp_headers):
                 errors += f"[{Context.COLLECTIONS}] /collections content-type header was not application/json"
 
             if not (self_link := link_by_rel(body.get("links", []), "self")):
@@ -722,6 +851,12 @@ def validate_collections(
                     f"[{Context.COLLECTIONS}] /collections does not have root link"
                 )
 
+            if collections_type := body.get("type"):
+                warnings += (
+                    f"[{Context.COLLECTIONS}] /collections entity has a field 'type: {collections_type}', "
+                    "but the STAC API entity schema does not define this field"
+                )
+
             if body.get("collections") is None:
                 errors += f"[{Context.COLLECTIONS}] /collections does not have 'collections' field"
 
@@ -745,11 +880,7 @@ def validate_collections(
             if not body:
                 errors += f"[{Context.COLLECTIONS}] {collection_url} body was empty"
             else:
-                if (
-                    not resp_headers
-                    or resp_headers.get("content-type", "").split(";")[0]
-                    != "application/json"
-                ):
+                if not resp_headers or not has_json_content_type(resp_headers):
                     errors += f"[{Context.COLLECTIONS}] {collection_url} content-type header was not application/json"
 
                 if not (self_link := link_by_rel(body.get("links", []), "self")):
@@ -910,9 +1041,6 @@ def validate_features(
                     if not link_by_rel(body.get("links", []), "root"):
                         errors += f"[{Context.FEATURES}] GET {collection_items_url} does not have root link"
 
-                    if not link_by_rel(body.get("links", []), "parent"):
-                        errors += f"[{Context.FEATURES}] GET {collection_items_url} does not have parent link"
-
                     stac_validate(collection_items_url, body, errors, Context.FEATURES)
 
                     item = next(iter(body.get("features", [])), None)
@@ -998,10 +1126,6 @@ def validate_features(
             errors=errors,
             r_session=r_session,
         )
-    else:
-        logger.info(
-            "Skipping STAC API - Features - Filter Extension conformance class."
-        )
 
 
 def validate_item_search(
@@ -1132,10 +1256,6 @@ def validate_item_search(
             errors=errors,
             r_session=r_session,
         )
-    else:
-        logger.info(
-            "Skipping STAC API - Item Search - Filter Extension conformance class."
-        )
 
 
 def validate_filter_queryables(
@@ -1238,10 +1358,6 @@ def validate_item_search_filter(
         logger.info(
             "Validating STAC API - Item Search - Filter Extension - CQL2-Text conformance class."
         )
-    else:
-        logger.info(
-            "Skipping STAC API - Item Search - Filter Extension - CQL2-Text conformance class."
-        )
 
     cql2_json_supported = (
         "http://www.opengis.net/spec/cql2/1.0/conf/cql2-json" in conforms_to
@@ -1254,10 +1370,6 @@ def validate_item_search_filter(
         logger.info(
             "Validating STAC API - Item Search - Filter Extension - CQL2-JSON conformance class."
         )
-    else:
-        logger.info(
-            "Skipping STAC API - Item Search - Filter Extension - CQL2-JSON conformance class."
-        )
 
     basic_cql2_supported = (
         "http://www.opengis.net/spec/cql2/1.0/conf/basic-cql2" in conforms_to
@@ -1270,10 +1382,6 @@ def validate_item_search_filter(
         logger.info(
             "Validating STAC API - Item Search - Filter Extension - Basic CQL2 conformance class."
         )
-    else:
-        logger.info(
-            "Skipping STAC API - Item Search - Filter Extension - Basic CQL2 conformance class."
-        )
 
     advanced_comparison_operators_supported = (
         "http://www.opengis.net/spec/cql2/1.0/conf/advanced-comparison-operators"
@@ -1284,10 +1392,6 @@ def validate_item_search_filter(
         logger.info(
             "Validating STAC API - Item Search - Filter Extension - Advanced Comparison Operators conformance class."
         )
-    else:
-        logger.info(
-            "Skipping STAC API - Item Search - Filter Extension - Advanced Comparison Operators conformance class."
-        )
 
     basic_spatial_operators_supported = (
         "http://www.opengis.net/spec/cql2/1.0/conf/basic-spatial-operators"
@@ -1298,10 +1402,6 @@ def validate_item_search_filter(
         logger.info(
             "Validating STAC API - Item Search - Filter Extension - Basic Spatial Operators conformance class."
         )
-    else:
-        logger.info(
-            "Skipping STAC API - Item Search - Filter Extension - Basic Spatial Operators conformance class."
-        )
 
     temporal_operators_supported = (
         "http://www.opengis.net/spec/cql2/1.0/conf/temporal-operators" in conforms_to
@@ -1311,10 +1411,6 @@ def validate_item_search_filter(
         logger.info(
             "Validating STAC API - Item Search - Filter Extension - Temporal Operators conformance class."
         )
-    else:
-        logger.info(
-            "Skipping STAC API - Item Search - Filter Extension - Temporal Operators conformance class."
-        )
 
     # todo: validate these
     # Spatial Operators: http://www.opengis.net/spec/cql2/1.0/conf/spatial-operators
@@ -1381,7 +1477,6 @@ def validate_item_search_filter(
             filter_jsons.append(cql2_json_not_like)
 
     if basic_spatial_operators_supported:
-
         if cql2_text_supported:
             filter_texts.append(cql2_text_s_intersects)
             filter_texts.append(cql2_text_ex_2(collection))
@@ -1731,13 +1826,14 @@ def validate_item_search_intersects(
         )
 
         if body:
-            if not len(body["features"]):
+            if not body.get("features"):
                 errors += f"[{Context.ITEM_SEARCH}] GET {search_url} Search result for intersects={geometry} returned no results"
-            if any(
-                not intersects_shape.intersects(shape(item["geometry"]))
-                for item in body["features"]
-            ):
-                errors += f"[{Context.ITEM_SEARCH}] GET {search_url} Search results for intersects={geometry} do not all intersect"
+            else:
+                if any(
+                    not intersects_shape.intersects(shape(item.get("geometry", None)))
+                    for item in body.get("features", [])
+                ):
+                    errors += f"[{Context.ITEM_SEARCH}] GET {search_url} Search results for intersects={geometry} do not all intersect"
 
     if Method.POST in methods:
         _, item_collection, _ = retrieve(
@@ -1748,11 +1844,12 @@ def validate_item_search_intersects(
             body={"collections": [collection], "intersects": geometry},
             r_session=r_session,
         )
-        if not (item_collection and len(item_collection["features"])):
+        if not item_collection or not item_collection.get("features"):
             errors += f"[{Context.ITEM_SEARCH}] POST Search result for intersects={geometry} returned no results"
-        for item in item_collection["features"]:  # type: ignore
-            if not intersects_shape.intersects(shape(item["geometry"])):
-                errors += f"[{Context.ITEM_SEARCH}] POST Search result for intersects={geometry}, does not intersect {item['geometry']}"
+        else:
+            for item in item_collection.get("features", []):
+                if not intersects_shape.intersects(shape(item.get("geometry"))):
+                    errors += f"[{Context.ITEM_SEARCH}] POST Search result for intersects={geometry}, does not intersect {item.get('geometry')}"
 
 
 def validate_item_search_bbox(
@@ -1773,7 +1870,7 @@ def validate_item_search_bbox(
     if Method.POST in methods:
         # Valid POST query
         _, body, resp_headers = retrieve(
-            Method.GET,
+            Method.POST,
             search_url,
             errors,
             Context.ITEM_SEARCH,
@@ -2115,7 +2212,8 @@ def validate_item_search_ids_does_not_override_all_other_params(
         )
     else:
         warnings += (
-            f"[{Context.ITEM_SEARCH}] GET Search with no parameters returned 0 results"
+            f"[{Context.ITEM_SEARCH}] GET Search within bbox=20,20,21,21 to validate ids does not override "
+            "all other parameters returned 0 results"
         )
 
 
